@@ -29,7 +29,7 @@ class RequestStatus:
         self._connection_acquired = AsyncEvent()
 
     async def wait_for_connection(
-        self, timeout: float = None
+        self, timeout: Optional[float] = None
     ) -> AsyncConnectionInterface:
         await self._connection_acquired.wait(timeout=timeout)
         assert self.connection is not None
@@ -43,16 +43,16 @@ class AsyncConnectionPool(AsyncRequestInterface):
 
     def __init__(
         self,
-        ssl_context: ssl.SSLContext = None,
+        ssl_context: Optional[ssl.SSLContext] = None,
         max_connections: Optional[int] = 10,
-        max_keepalive_connections: int = None,
-        keepalive_expiry: float = None,
+        max_keepalive_connections: Optional[int] = None,
+        keepalive_expiry: Optional[float] = None,
         http1: bool = True,
         http2: bool = False,
         retries: int = 0,
-        local_address: str = None,
-        uds: str = None,
-        network_backend: AsyncNetworkBackend = None,
+        local_address: Optional[str] = None,
+        uds: Optional[str] = None,
+        network_backend: Optional[AsyncNetworkBackend] = None,
     ) -> None:
         """
         A connection pool for making HTTP requests.
@@ -214,16 +214,24 @@ class AsyncConnectionPool(AsyncRequestInterface):
             )
 
         status = RequestStatus(request)
+        self._requests.append(status)
 
         async with self._pool_lock:
-            self._requests.append(status)
             await self._close_expired_connections()
             await self._attempt_to_acquire_connection(status)
 
         while True:
             timeouts = request.extensions.get("timeout", {})
             timeout = timeouts.get("pool", None)
-            connection = await status.wait_for_connection(timeout=timeout)
+            try:
+                connection = await status.wait_for_connection(timeout=timeout)
+            except BaseException as exc:
+                # If we timeout here, or if the task is cancelled, then make
+                # sure to remove the request from the queue before bubbling
+                # up the exception.
+                self._requests.remove(status)
+                raise exc
+
             try:
                 response = await connection.handle_async_request(request)
             except ConnectionNotAvailable:
@@ -265,10 +273,11 @@ class AsyncConnectionPool(AsyncRequestInterface):
         assert status.connection is not None
         connection = status.connection
 
+        if status in self._requests:
+            self._requests.remove(status)
+
         async with self._pool_lock:
             # Update the state of the connection pool.
-            if status in self._requests:
-                self._requests.remove(status)
 
             if connection.is_closed() and connection in self._pool:
                 self._pool.remove(connection)
@@ -310,9 +319,9 @@ class AsyncConnectionPool(AsyncRequestInterface):
 
     async def __aexit__(
         self,
-        exc_type: Type[BaseException] = None,
-        exc_value: BaseException = None,
-        traceback: TracebackType = None,
+        exc_type: Optional[Type[BaseException]] = None,
+        exc_value: Optional[BaseException] = None,
+        traceback: Optional[TracebackType] = None,
     ) -> None:
         await self.aclose()
 
